Files
Jungfraujoch/image_analysis/bragg_prediction/BraggPredictionRotGPU.cu
T
leonarski_fandjungfrau 4dc2534dbf
Build Packages / build:rpm (rocky9_sls9) (push) Successful in 18m57s
Build Packages / Unit tests (push) Skipped
Build Packages / build:windows:nocuda (push) Successful in 16m55s
Build Packages / build:windows:cuda (push) Successful in 18m48s
Build Packages / build:viewer-tgz:cpu (push) Successful in 13m10s
Build Packages / build:viewer-tgz:cuda (push) Successful in 14m45s
Build Packages / build:rpm (rocky8_nocuda) (push) Successful in 22m23s
Build Packages / build:rpm (rocky9_nocuda) (push) Successful in 20m12s
Build Packages / build:rpm (ubuntu2204_nocuda) (push) Successful in 23m7s
Build Packages / build:rpm (ubuntu2404_nocuda) (push) Successful in 20m43s
Build Packages / build:rpm (rocky8_sls9) (push) Successful in 23m9s
Build Packages / XDS test (durin plugin) (push) Successful in 12m26s
Build Packages / build:rpm (rocky9) (push) Successful in 24m58s
Build Packages / Generate python client (push) Successful in 50s
Build Packages / build:rpm (ubuntu2404) (push) Successful in 23m20s
Build Packages / Create release (push) Skipped
Build Packages / XDS test (JFJoch plugin) (push) Successful in 12m37s
Build Packages / build:rpm (rocky8) (push) Successful in 27m58s
Build Packages / build:rpm (ubuntu2204) (push) Successful in 25m38s
Build Packages / Build documentation (push) Successful in 59s
Build Packages / DIALS test (push) Successful in 23m16s
Build Packages / XDS test (neggia plugin) (push) Successful in 6m38s
v1.0.0.rc-162 (#72)
**Files written by Jungfraujoch now import correctly in DIALS, XDS and pyFAI.** A tilted detector, a grid scan, a still recorded at a goniometer position, and saturated or unreadable pixels were each described in a way that a third-party program acted on wrongly. If you process Jungfraujoch data outside Jungfraujoch, prefer this release to any earlier one.

* HDF5: the detector tilt (`rot1`/`rot2`/`rot3`) is exported correctly in the NXmx transformation chain; untilted geometries are unaffected.
* HDF5: a still recorded at a goniometer position is no longer read back as a single image, and a grid scan records a stationary spindle so a program that requires a rotation axis can open it.
* HDF5: the sample transformation chain is written in mounting order, with a Smargon head position told apart from the spindle, one entry per image, `module_offset` as a float unit vector, and `offset_units` on every offset.
* HDF5: saturated, underloaded and unreadable pixels are described so a downstream program masks them - `saturation_value`, `underload_value`, `error_value` and `bit_depth_readout` are written correctly, and a data file missing next to a VDS master reads as the error marker rather than as zero counts.
* HDF5: the rotation axis is read back under whatever name it carries, and `mirror_y` records whether the assembled image is mirrored in Y relative to the detector's raw readout.
* A grid scan and a goniometer axis can both be set; they are no longer alternatives.
* `images_per_file` is chosen from the acquisition when it is not given: a rotation sweep of at most 20000 images goes into a single data file, a grid scan splits on whole fast-axis rows, and stills and serial keep 1000.
* The writer refuses a stream whose start message declares a different pixel format than its images carry, and a DECTRIS detector sending signed images is no longer declared unsigned.
* The image stream can carry the sample transformation chain (`transformations`, in the END message); a producer that does not send it gets the same chain built by the writer.
* rugnux: fixing the space group with `-S` no longer prevents the lattice from being found - a lattice indexed in a different setting is reindexed into that group's own setting, and a run whose crystal does not have that group's lattice stops and names the cell it indexed as, rather than reporting statistics that cannot describe it.
* rugnux: the per-image resolution estimate now predicts the resolution the merged data reach rather than the highest-resolution spot found, and is reported as `SPOT_RESOLUTION_ESTIMATE`.
* rugnux: two runs of the same command on the same images produce the same merged intensities; the azimuthal profile written alongside them is not yet reproducible in the same way.
* rugnux: the offline lattice refinement is bounded by iterations rather than by a wall clock, so a loaded machine can no longer refine to a different lattice; a live acquisition keeps its real-time bound.
* rugnux: the detector-frame modulation correction is fitted on a grid spanning the detector, so whether it is applied no longer depends on how far integration reached.
* rugnux: the geometry pre-pass no longer writes `<prefix>_01.mtz`, `_01.cif`, `_01.hkl` and `_01_image.dat`; the refined second pass writes those files under `<prefix>`, and that is the result to use.
* rugnux: `_process.h5` describes the pixel format of the images it links to, and is written on a thread of its own.
* rugnux: the detector geometry is also logged in XDS's convention (`ORGX`/`ORGY`, detector axis vectors, rotation axis), so it can be compared with an XDS refinement.
* rugnux: an image integrated in pyFAI through the `.poni` file written by `--mode calibration` comes out with the correct azimuth, and the file declares pyFAI's `orientation`, which needs pyFAI 2024.01 or newer. Radial integration is unchanged.
* rugnux: a rotation run is substantially faster throughout - beam-stop detection, first-pass indexing, geometry refinement, integration, scaling and merging - and observations outside the scaling resolution range are dropped as they are ingested. The refined geometry, the space group chosen and the merged statistics are unchanged.
* Faster spot finding and indexing, on the broker as well as in rugnux; the spots found and the lattices indexed are unchanged.
* A run reserves substantially less GPU memory: nothing is allocated for buffers that are never read, and a worker builds only the engines it uses.
* rugnux: with `-N` left at its default the per-image loop of `--mode mx` uses at most 16 workers per GPU, rather than one per hardware thread; an explicit `-N` is obeyed as given.
* CUDA 12 builds now contain device code for Volta, so the RHEL 8 packages and the portable Linux `.tgz` run on a V100; the CUDA 13 artefacts (RHEL 9, Ubuntu, Windows) remain Turing and newer.
* The build resolves a single Eigen for the whole project, and refuses to configure if Ceres picks up a different one; a build that mixed two Eigen versions was undefined behaviour and crashed at -O2.
* Documentation: a security page, and the supported GPU generations and minimum NVIDIA driver version of every released artefact.

**Breaking change to OpenAPI** - regenerate the client (`jfjoch-client` 1.0.0-rc.162, `frontend/src/client`):
* `dataset_settings.images_per_file` is no longer `default: 1000` and no longer accepts `0`; it is optional, and its minimum is 1. A client sending `0` (previously "one file for the whole run") is now rejected - omit the field instead, which for a rotation sweep gives the same single file.
* `file_writer_format` now defaults to `NXmxVDS`, matching the server's own default and the layout recommended for DIALS, XDS and CrystFEL. A generated client that fills in schema defaults and does not set the format explicitly will write VDS masters where it previously wrote legacy ones; set `NXmxLegacy` explicitly to keep them.

---------

Co-authored-by: jungfrau <jungfrau@mx-aare-test.psi.ch>
Reviewed-on: #72
Co-authored-by: Filip Leonarski <filip.leonarski@psi.ch>
2026-08-25 08:21:39 +02:00

336 lines
14 KiB
Plaintext

// SPDX-FileCopyrightText: 2025 Filip Leonarski, Paul Scherrer Institute <filip.leonarski@psi.ch>
// SPDX-License-Identifier: GPL-3.0-only
#include <algorithm>
#include "../../common/JFJochMath.h"
#include "BraggPredictionRotGPU.h"
#ifdef JFJOCH_USE_CUDA
#include "../indexing/CUDAMemHelpers.h"
#include <cuda_runtime.h>
#include <cmath>
namespace {
__host__ __device__ inline bool is_odd(int v) { return (v & 1) != 0; }
__host__ __device__ inline void cross3(float ax, float ay, float az,
float bx, float by, float bz,
float &cx, float &cy, float &cz) {
cx = ay * bz - az * by;
cy = az * bx - ax * bz;
cz = ax * by - ay * bx;
}
__host__ __device__ inline float dot3(float ax, float ay, float az,
float bx, float by, float bz) {
return ax * bx + ay * by + az * bz;
}
__host__ __device__ inline void normalize3(float &x, float &y, float &z) {
float len = sqrtf(x * x + y * y + z * z);
if (len < 1e-12f) { x = 0.0f; y = 0.0f; z = 0.0f; return; }
float inv = 1.0f / len;
x *= inv; y *= inv; z *= inv;
}
__device__ inline int compute_reflections_rot(const KernelConstsRot &C, int h, int k, int l, Reflection out[2]) {
if (h == 0 && k == 0 && l == 0)
return 0;
switch (C.centering) {
case 'I':
if (is_odd(h + k + l)) return false;
break;
case 'A':
if (is_odd(k + l)) return false;
break;
case 'B':
if (is_odd(h + l)) return false;
break;
case 'C':
if (is_odd(h + k)) return false;
break;
case 'F':
if (is_odd(h + k) || is_odd(h + l) || is_odd(k + l)) return false;
break;
case 'R': {
int mod = (-h + k + l) % 3;
if (mod < 0) mod += 3;
if (mod != 0) return false;
break;
}
default:
break;
}
// p0 = A* h + B* k + C* l
float p0x = C.Astar.x * h + C.Bstar.x * k + C.Cstar.x * l;
float p0y = C.Astar.y * h + C.Bstar.y * k + C.Cstar.y * l;
float p0z = C.Astar.z * h + C.Bstar.z * k + C.Cstar.z * l;
float p0_sq = p0x * p0x + p0y * p0y + p0z * p0z;
if (p0_sq <= 0.0f || p0_sq > C.one_over_dmax_sq)
return 0;
float p0_m1 = p0x * C.m1.x + p0y * C.m1.y + p0z * C.m1.z;
float p0_m2 = p0x * C.m2.x + p0y * C.m2.y + p0z * C.m2.z;
float p0_m3 = p0x * C.m3.x + p0y * C.m3.y + p0z * C.m3.z;
float rho_sq = p0_sq - (p0_m2 * p0_m2);
float p_m3 = (-p0_sq / 2.0f - p0_m2 * C.m2_S0) / C.m3_S0;
float p_m2 = p0_m2;
if (rho_sq < p_m3 * p_m3) return 0;
if (p0_sq > 4.0f * dot3(C.S0.x, C.S0.y, C.S0.z, C.S0.x, C.S0.y, C.S0.z)) return 0;
float p_m1_pos = sqrtf(rho_sq - p_m3 * p_m3);
float p_m1_arr[2] = {p_m1_pos, -p_m1_pos};
// Effective rocking width: mosaicity broadened by the bandwidth term (consistent with CPU),
// dtheta = (dlambda/lambda) tan(theta_B) in quadrature with sigma_M, zeta-free.
float mos_eff_rad = C.mos_angle_rad;
if (C.bandwidth_sigma > 0.0f) {
float sin_theta = C.half_wavelength_A * sqrtf(p0_sq);
float dphi_bw = C.bandwidth_sigma * sin_theta / sqrtf(1.0f - sin_theta * sin_theta);
mos_eff_rad = sqrtf(C.mos_angle_rad * C.mos_angle_rad + dphi_bw * dphi_bw);
}
int count = 0;
for (int idx = 0; idx < 2; ++idx) {
float p_m1 = p_m1_arr[idx];
float cosphi = (p_m1 * p0_m1 + p_m3 * p0_m3) / rho_sq;
float sinphi = (p_m1 * p0_m3 - p_m3 * p0_m1) / rho_sq;
float px = C.m1.x * p_m1 + C.m2.x * p_m2 + C.m3.x * p_m3;
float py = C.m1.y * p_m1 + C.m2.y * p_m2 + C.m3.y * p_m3;
float pz = C.m1.z * p_m1 + C.m2.z * p_m2 + C.m3.z * p_m3;
float Sx = C.S0.x + px;
float Sy = C.S0.y + py;
float Sz = C.S0.z + pz;
float phi = -1.0f * atan2f(sinphi, cosphi);
// e1 = normalize(S x S0) - direction perpendicular to both S and S0
float e1x, e1y, e1z;
cross3(Sx, Sy, Sz, C.S0.x, C.S0.y, C.S0.z, e1x, e1y, e1z);
normalize3(e1x, e1y, e1z);
// zeta = |m2 · e1| - the "lorentz-like" geometric factor for partiality
float zeta_abs = fabsf(dot3(C.m2.x, C.m2.y, C.m2.z, e1x, e1y, e1z));
// Check min_zeta threshold (consistent with CPU)
if (zeta_abs < C.min_zeta)
continue;
// epsilon3 cutoff check (consistent with CPU, Kabsch formulation): measured against the
// frame's EDGE, not its centre, or a reflection whose rocking curve overlaps the exposure
// is rejected because its exact diffracting condition falls outside it - and since the
// nearest frame centre is at most half a wedge away, it is then rejected on every frame
// and lost entirely. See the CPU engine for the regime that reaches.
float epsilon3 = fabsf(phi * zeta_abs) - 0.5f * C.wedge_angle_rad * zeta_abs;
if (epsilon3 > C.mosaicity_multiplier * mos_eff_rad)
continue;
float cx, cy, cz;
cross3(Sx, Sy, Sz, C.S0.x, C.S0.y, C.S0.z, cx, cy, cz);
// Reciprocal Lorentz (Kabsch 2010): |m2 . (S x S0)| / (|S| |S0|) = zeta * sin(2theta).
// Dividing by the scalar product S.S0 = |S||S0|cos(2theta) would add a spurious
// 1/cos(2theta) to the absolute scale.
float S_len = sqrtf(dot3(Sx, Sy, Sz, Sx, Sy, Sz));
float S0_len = sqrtf(dot3(C.S0.x, C.S0.y, C.S0.z, C.S0.x, C.S0.y, C.S0.z));
float lorentz = fabsf(dot3(C.m2.x, C.m2.y, C.m2.z, cx, cy, cz)) / (S_len * S0_len);
// Partiality calculation (Kabsch formulation)
// c1 = sqrt(2) * sigma / zeta, where sigma = mosaicity
float c1 = zeta_abs / (sqrtf(2.0f) * mos_eff_rad);
float half_wedge = C.wedge_angle_rad / 2.0f;
float partiality = (erff((phi + half_wedge) * c1)
- erff((phi - half_wedge) * c1)) / 2.0f;
// Use S (rotated) for projection
float Srx = C.rot[0] * Sx + C.rot[1] * Sy + C.rot[2] * Sz;
float Sry = C.rot[3] * Sx + C.rot[4] * Sy + C.rot[5] * Sz;
float Srz = C.rot[6] * Sx + C.rot[7] * Sy + C.rot[8] * Sz;
if (Srz <= 0.0f) continue;
float coeff = C.coeff_const / Srz;
float x = C.beam_x + Srx * coeff;
float y = C.beam_y + Sry * coeff;
if (x < 0.0f || x >= C.det_width_pxl || y < 0.0f || y >= C.det_height_pxl)
continue;
float dist_ewald = fabsf(sqrtf(Sx * Sx + Sy * Sy + Sz * Sz) - C.one_over_wavelength);
out[count].h = h;
out[count].k = k;
out[count].l = l;
out[count].delta_phi_deg = phi * 180.0 / PI;
out[count].predicted_x = x;
out[count].predicted_y = y;
out[count].observed_x = NAN;
out[count].observed_y = NAN;
out[count].d = 1.0f / sqrtf(p0_sq);
out[count].dist_ewald = dist_ewald;
out[count].rlp = lorentz;
out[count].partiality = partiality;
out[count].zeta = zeta_abs;
out[count].image_scale_corr = lorentz / partiality;
count++;
}
return count;
}
__global__ void bragg_rot_kernel_3d(const KernelConstsRot *__restrict__ kc,
int max_h, int max_k, int max_l,
int max_reflections,
Reflection *__restrict__ out,
int *__restrict__ counter) {
int hi = blockIdx.x * blockDim.x + threadIdx.x;
int ki = blockIdx.y * blockDim.y + threadIdx.y;
int li = blockIdx.z * blockDim.z + threadIdx.z;
if (hi > 2 * max_h || ki > 2 * max_k || li > 2 * max_l) return;
int h = hi - max_h;
int k = ki - max_k;
int l = li - max_l;
Reflection r[2];
int n = compute_reflections_rot(*kc, h, k, l, r);
for (int i = 0; i < n; ++i) {
// Do NOT clamp the counter back down on overflow: it then saturates at the capacity and the
// host cannot tell a full buffer from an overflowing one. Let it count the true total.
const int pos = atomicAdd(counter, 1);
if (pos < max_reflections)
out[pos] = r[i];
}
}
inline KernelConstsRot BuildKernelConstsRot(const DiffractionExperiment &experiment,
const CrystalLattice &lattice,
const BraggPredictionSettings &settings) {
KernelConstsRot kc{};
auto geom = experiment.GetDiffractionGeometry();
kc.det_width_pxl = static_cast<float>(experiment.GetXPixelsNum());
kc.det_height_pxl = static_cast<float>(experiment.GetYPixelsNum());
kc.beam_x = geom.GetBeamX_pxl();
kc.beam_y = geom.GetBeamY_pxl();
kc.coeff_const = geom.GetDetectorDistance_mm() / geom.GetPixelSize_mm();
float one_over_dmax = 1.0f / settings.high_res_A;
kc.one_over_dmax_sq = one_over_dmax * one_over_dmax;
kc.one_over_wavelength = 1.0f / geom.GetWavelength_A();
// Store mosaicity and wedge in radians for partiality calculation
kc.mos_angle_rad = settings.mosaicity_deg * static_cast<float>(PI) / 180.0f;
kc.wedge_angle_rad = settings.wedge_deg * static_cast<float>(PI) / 180.0f;
kc.min_zeta = settings.min_zeta;
kc.mosaicity_multiplier = settings.mosaicity_multiplier;
kc.bandwidth_sigma = settings.bandwidth_sigma;
kc.half_wavelength_A = geom.GetWavelength_A() / 2.0f;
kc.Astar = lattice.Astar();
kc.Bstar = lattice.Bstar();
kc.Cstar = lattice.Cstar();
kc.S0 = geom.GetScatteringVector();
auto rotT = geom.GetPoniRotMatrix().transpose().arr();
for (int i = 0; i < 9; ++i) kc.rot[i] = rotT[i];
kc.centering = settings.centering;
return kc;
}
inline void BuildGoniometerBasis(const DiffractionExperiment &experiment, KernelConstsRot &kc) {
const auto gon_opt = experiment.GetGoniometer();
if (!gon_opt.has_value())
throw JFJochException(JFJochExceptionCategory::InputParameterInvalid,
"BraggPredictionRotationGPU requires a goniometer axis");
const GoniometerAxis &gon = *gon_opt;
// m2 = normalize(axis)
float m2x = gon.GetAxis().x;
float m2y = gon.GetAxis().y;
float m2z = gon.GetAxis().z;
normalize3(m2x, m2y, m2z);
// m1 = normalize(m2 x S0)
float m1x, m1y, m1z;
cross3(m2x, m2y, m2z, kc.S0.x, kc.S0.y, kc.S0.z, m1x, m1y, m1z);
normalize3(m1x, m1y, m1z);
// m3 = normalize(m1 x m2)
float m3x, m3y, m3z;
cross3(m1x, m1y, m1z, m2x, m2y, m2z, m3x, m3y, m3z);
normalize3(m3x, m3y, m3z);
kc.m1 = Coord(m1x, m1y, m1z);
kc.m2 = Coord(m2x, m2y, m2z);
kc.m3 = Coord(m3x, m3y, m3z);
kc.m2_S0 = dot3(m2x, m2y, m2z, kc.S0.x, kc.S0.y, kc.S0.z);
kc.m3_S0 = dot3(m3x, m3y, m3z, kc.S0.x, kc.S0.y, kc.S0.z);
}
} // namespace
BraggPredictionRotGPU::BraggPredictionRotGPU(int max_reflections)
: BraggPrediction(max_reflections),
reg_out(reflections), d_out(max_reflections),
dK(1), d_count(1), h_count(1) {
}
// The host buffer is page-locked and the device buffer is sized to match it, so both are rebuilt.
void BraggPredictionRotGPU::GrowCapacity(int count) {
reg_out = CudaRegisteredVector<Reflection>(); // unregister before the vector reallocates
BraggPrediction::GrowCapacity(count);
reg_out = CudaRegisteredVector<Reflection>(reflections);
d_out = CudaDevicePtr<Reflection>(count);
}
int BraggPredictionRotGPU::Calc(const DiffractionExperiment &experiment,
const CrystalLattice &lattice,
const BraggPredictionSettings &settings) {
KernelConstsRot hK = BuildKernelConstsRot(experiment, lattice, settings);
BuildGoniometerBasis(experiment, hK);
cudaMemcpyAsync(dK, &hK, sizeof(KernelConstsRot), cudaMemcpyHostToDevice, stream);
cudaMemsetAsync(d_count, 0, sizeof(int), stream);
// Inclusive on both ends, matching the kernel's own bounds and the CPU loops (-max_i .. +max_i).
dim3 block(8, 8, 8);
dim3 grid((2 * settings.max_h + 1 + block.x - 1) / block.x,
(2 * settings.max_k + 1 + block.y - 1) / block.y,
(2 * settings.max_l + 1 + block.z - 1) / block.z);
bragg_rot_kernel_3d<<<grid, block, 0, stream>>>(dK, settings.max_h, settings.max_k, settings.max_l, max_reflections, d_out, d_count);
cudaMemcpyAsync(h_count, d_count, sizeof(int), cudaMemcpyDeviceToHost, stream);
cudaStreamSynchronize(stream);
int count = *h_count.get();
if (count > max_reflections) {
// The buffer holds an arbitrary subset of what was predicted (whichever slots the atomics
// reached first), so it cannot be used. Grow to fit and predict again; the buffer stays grown,
// so a run pays for this a handful of times at most.
GrowCapacity(count);
cudaMemsetAsync(d_count, 0, sizeof(int), stream);
bragg_rot_kernel_3d<<<grid, block, 0, stream>>>(dK, settings.max_h, settings.max_k, settings.max_l, max_reflections, d_out, d_count);
cudaMemcpyAsync(h_count, d_count, sizeof(int), cudaMemcpyDeviceToHost, stream);
cudaStreamSynchronize(stream);
count = std::min(*h_count.get(), max_reflections);
}
if (count == 0) return 0;
cudaMemcpyAsync(reflections.data(), d_out, sizeof(Reflection) * count, cudaMemcpyDeviceToHost, stream);
cudaStreamSynchronize(stream);
OrderOutput(count);
return TruncateToOutput(count);
}
#endif